﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Windows.Forms;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding.Views;
using HeuristicLab.EvolutionTracking;
using HeuristicLab.EvolutionTracking.Views;
using HeuristicLab.MainForm;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic.Views {
  [View("SymbolicDataAnalysisGenealogyGraphView")]
  [Content(typeof(IGenealogyGraph<ISymbolicExpressionTree>), IsDefaultView = true)]
  public partial class SymbolicDataAnalysisGenealogyGraphView : SymbolicDataAnalysisGenealogyGraphViewDesignable {
    private readonly ISymbolicExpressionTreeNodeEqualityComparer comparer;

    private ISymbolicExpressionTreeNode selectedSubtree;
    private ISymbolicExpressionTreeNode SelectedSubtree {
      get { return selectedSubtree; }
      set {
        if (value == null || selectedSubtree == value) return;
        selectedSubtree = value;
        ClonedSubtree = (ISymbolicExpressionTreeNode)selectedSubtree.Clone();
      }
    }
    // the clone is necessary in order to deselect parts of a tree (by removing subtrees) 
    // and do "high level" matching of the remaining part
    private ISymbolicExpressionTreeNode ClonedSubtree { get; set; }

    // we use the symbolic expression tree chart to call the drawing routines
    // and highlight relevant tree parts
    public SymbolicExpressionTreeChart SymbolicExpressionTreeChart {
      get {
        var view = (GraphicalSymbolicExpressionTreeView)base.viewHost.ActiveView;
        return view == null ? null : view.SymbolicExpressionTreeChart;
      }
    }

    public SymbolicDataAnalysisGenealogyGraphView() {
      InitializeComponent();
      comparer = new SymbolicExpressionTreeNodeEqualityComparer();
      viewHost.ViewType = typeof(GraphicalSymbolicExpressionTreeView);
    }
    #region event handlers
    protected override void OnContentChanged() {
      base.OnContentChanged();
      if (Content != null && Content != genealogyGraphChart.GenealogyGraph) {
        genealogyGraphChart.GenealogyGraph = Content;
      }
    }

    public override void graphChart_GenealogyGraphNodeClicked(object sender, MouseEventArgs args) {
      base.graphChart_GenealogyGraphNodeClicked(sender, args);
      var visualNode = (VisualGenealogyGraphNode)sender;
      var graphNode = (IGenealogyGraphNode<ISymbolicExpressionTree>)visualNode.Data;

      nodeQualityLabel.Text = String.Format("{0:0.000}", graphNode.Quality);
      nodeRankLabel.Text = String.Format("{0:0.0}", graphNode.Rank);
      nodeDegreeLabel.Text = String.Format("{0} / {1}", graphNode.InDegree, graphNode.OutDegree);
      nodeWeightLabel.Text = String.Format("{0:0.00}", graphNode.Weight);

      int colorCount = ColorGradient.Colors.Count - 1;
      const int alpha = 100;
      // calculate max only in the current generation
      var nodes = graphNode.Data.IterateNodesPrefix().Skip(1).ToList();
      var max = Content.GetByRank(graphNode.Rank).Max(x => ((IGenealogyGraphNode<ISymbolicExpressionTree>)x).Data.IterateNodesPrefix().Max(n => n.NodeWeight));
      if (max.IsAlmost(0)) max = 1;
      foreach (var n in nodes) {
        var visualSymbolicExpressionTreeNode = SymbolicExpressionTreeChart.GetVisualSymbolicExpressionTreeNode(n);
        visualSymbolicExpressionTreeNode.ToolTip += Environment.NewLine + String.Format("Weight: {0:0}", n.NodeWeight);
        var i = (int)Math.Round(n.NodeWeight / max * colorCount);
        var fillColor = Color.FromArgb(alpha, ColorGradient.Colors[i]);
        treeChart_HighlightSubtree(n, null, fillColor);
      }

      if (!graphNode.InArcs.Any()) return;

      if (openNew_CheckBox.Checked) {
        // get the ancestors into a new view
        var cloner = new Cloner();
        // clone arcs and vertices and use them to create a new graph 
        // that will include just the individual and its ancestors
        var graph = new GenealogyGraph<ISymbolicExpressionTree>();
        var ancestors = new[] { graphNode }.Concat(graphNode.Ancestors).ToList();
        graph.AddVertices(ancestors.Select(cloner.Clone));
        graph.AddArcs(ancestors.SelectMany(x => x.InArcs).Select(cloner.Clone));
        MainFormManager.MainForm.ShowContent(graph);
      }

      var data = graphNode.InArcs.Last().Data;
      var fragment = data as IFragment<ISymbolicExpressionTreeNode>;
      var td = data as TraceData;

      if (fragment != null) {
        // highlight the fragment with a blue line color
        foreach (var n in graphNode.Data.NodeAt(fragment.Index1).IterateNodesPrefix()) {
          treeChart_HighlightSubtree(n, Color.CornflowerBlue);
        }
      } else if (td != null) {
        // highlight traced subtree and fragment
        treeChart_HighlightSubtree(nodes[td.LastSubtreeIndex], Color.Orange);
        treeChart_HighlightSubtree(nodes[td.LastFragmentIndex], Color.CornflowerBlue);
      }
    }

    #region specific symbolic expression tree node click actions
    private void OnTreeNodeLeftClicked(ISymbolicExpressionTreeNode node) {
      SelectedSubtree = node;
      bool trace = genealogyGraphChart.TraceFragments;

      // check whether we are in 'trace' or 'match' mode
      if (trace) {
        // perform fragment tracing
        var graphNode = (IGenealogyGraphNode<ISymbolicExpressionTree>)genealogyGraphChart.SelectedGraphNode;
        var subtreeIndex = graphNode.Data.IterateNodesPrefix().ToList().IndexOf(node);
        var traceGraph = TraceCalculator.TraceSubtree(graphNode, subtreeIndex, updateVertexWeights: false, updateSubtreeWeights: false, cacheTraceNodes: true);
        genealogyGraphChart.UpdateEnabled = false;
        genealogyGraphChart.ClearPrimitives(); // clear everything
        var rankMaximums = new Dictionary<double, double>();
        // fill each vertex in the graph with a grayscale color according to average subtree weight
        foreach (var r in Content.Ranks) {
          var vertices = r.Value.Cast<IGenealogyGraphNode<ISymbolicExpressionTree>>().ToList();
          var averages = vertices.Select(x => x.Data.IterateNodesPrefix().Average(n => n.NodeWeight)).ToList();
          var max = averages.Max();
          rankMaximums.Add(r.Key, max);
          if (max.IsAlmost(0.0)) continue;
          int colorCount = ColorGradient.GrayscaledColors.Count - 1;
          for (int i = 0; i < vertices.Count; ++i) {
            var color = ColorGradient.GrayscaledColors[(int)Math.Round(averages[i] / max * colorCount)];
            var vNode = genealogyGraphChart.GetMappedNode(vertices[i]);
            vNode.Brush = new SolidBrush(color);
          }
        }
        // fill vertices that are part of the trace with a rgb color according to average subtree weight
        if (traceGraph.Vertices.Any()) {
          var ranks = traceGraph.Vertices.Select(x => Content.GetByContent(x.Data)).GroupBy(x => x.Rank);
          int colorCount = ColorGradient.Colors.Count - 1;
          foreach (var g in ranks) {
            var vertices = g.ToList();
            var averages = vertices.Select(x => x.Data.IterateNodesPrefix().Average(n => n.NodeWeight)).ToList();
            double max = rankMaximums[g.Key];
            if (max.IsAlmost(0.0)) continue;
            for (int i = 0; i < vertices.Count; ++i) {
              var vNode = genealogyGraphChart.GetMappedNode(vertices[i]);
              // use a grayscale gradient
              var color = ColorGradient.Colors[(int)Math.Round(averages[i] / max * colorCount)];
              vNode.Brush = new SolidBrush(color);
            }
          }

          if (openNew_CheckBox.Checked) {
            MainFormManager.MainForm.ShowContent(traceGraph); // display the fragment graph on the screen
          }
        }
        genealogyGraphChart.UpdateEnabled = true;
        genealogyGraphChart.EnforceUpdate();
      } else {
        // perform matching like it was done before
        // currently there is no possibility to specify the subtree matching criteria
        var trees = Content.Vertices.Select(x => x.Data);
        var matchingTrees = trees.Where(x => x.Root.ContainsSubtree(node, comparer));

        var matchingVertices = matchingTrees.Select(x => Content.GetByContent(x));
        graphChart_HighlightMatchingVertices(matchingVertices);
      }
    }

    private void OnTreeNodeMiddleClicked(ISymbolicExpressionTreeNode node) {
      var index = SelectedSubtree.IterateNodesPrefix().ToList().IndexOf(node);
      if (index > 0) {
        var s = ClonedSubtree.NodeAt(index);
        s.Parent.RemoveSubtree(s.Parent.IndexOfSubtree(s));

        var trees = Content.Vertices.Select(x => x.Data);
        var matchingTrees = trees.Where(x => x.Root.ContainsSubtree(ClonedSubtree, comparer));

        var matchingVertices = matchingTrees.Select(x => Content.GetByContent(x));
        treeChart_HighlightSubtree(node, Color.Black, Color.White);
        graphChart_HighlightMatchingVertices(matchingVertices);
      }
    }
    #endregion

    public void treeChart_SymbolicExpressionTreeNodeClicked(object sender, MouseEventArgs args) {
      var visualNode = (VisualTreeNode<ISymbolicExpressionTreeNode>)sender;
      var subtree = visualNode.Content;
      // highlight the selected subtree inside the displayed tree on the right hand side
      treeChart_ClearColors();
      treeChart_HighlightSubtree(subtree, Color.Black, Color.RoyalBlue);

      switch (args.Button) {
        case MouseButtons.Left: {
            OnTreeNodeLeftClicked(subtree);
            break;
          }
        case MouseButtons.Middle: {
            OnTreeNodeMiddleClicked(subtree);
            break;
          }
      }
    }
    #endregion

    protected override void RegisterContentEvents() {
      base.RegisterContentEvents();
      if (SymbolicExpressionTreeChart != null)
        SymbolicExpressionTreeChart.SymbolicExpressionTreeNodeClicked += treeChart_SymbolicExpressionTreeNodeClicked;
    }

    protected override void DeregisterContentEvents() {
      if (SymbolicExpressionTreeChart != null)
        SymbolicExpressionTreeChart.SymbolicExpressionTreeNodeClicked -= treeChart_SymbolicExpressionTreeNodeClicked;
      base.DeregisterContentEvents();
    }

    private void graphChart_HighlightMatchingVertices(IEnumerable<IGenealogyGraphNode> vertices) {
      genealogyGraphChart.Chart.UpdateEnabled = false;
      genealogyGraphChart.ClearPrimitives();
      genealogyGraphChart.HighlightNodes(vertices);
      genealogyGraphChart.Chart.UpdateEnabled = true;
      genealogyGraphChart.Chart.EnforceUpdate();
    }

    private void treeChart_ClearColors() {
      foreach (var node in SymbolicExpressionTreeChart.Tree.IterateNodesPrefix()) {
        var visualNode = SymbolicExpressionTreeChart.GetVisualSymbolicExpressionTreeNode(node);
        if (visualNode != null) {
          visualNode.LineColor = Color.Black;
          visualNode.FillColor = Color.Transparent;
        }
      }
      SymbolicExpressionTreeChart.RepaintNodes();
    }

    private void treeChart_HighlightSubtree(ISymbolicExpressionTreeNode subtree, Color? lineColor = null, Color? fillColor = null) {
      if (lineColor == null && fillColor == null)
        return;

      foreach (var s in subtree.IterateNodesPrefix()) {
        var visualNode = SymbolicExpressionTreeChart.GetVisualSymbolicExpressionTreeNode(s);
        if (lineColor != null) {
          visualNode.LineColor = (Color)lineColor;
          foreach (var c in s.Subtrees) {
            var visualArc = SymbolicExpressionTreeChart.GetVisualSymbolicExpressionTreeNodeConnection(s, c);
            visualArc.LineColor = (Color)lineColor;
          }
        }
        if (fillColor != null)
          visualNode.FillColor = (Color)fillColor;
      }
      SymbolicExpressionTreeChart.RepaintNodes();
    }

    #region navigate the genealogy / trace graph
    private void navigateLeftButton_Click(object sender, EventArgs e) {
      var graphNode = (IGenealogyGraphNode<ISymbolicExpressionTree>)genealogyGraphChart.SelectedGraphNode;
      var inArcs = (List<IArc>)((IVertex)graphNode).InArcs;
      if (inArcs.Count > 0) {
        genealogyGraphChart.SelectedGraphNode = (IGenealogyGraphNode<ISymbolicExpressionTree>)inArcs[0].Source;
      }
    }

    private void navigateRightButton_Click(object sender, EventArgs e) {
      var graphNode = (IGenealogyGraphNode<ISymbolicExpressionTree>)genealogyGraphChart.SelectedGraphNode;
      var inArcs = (List<IArc>)((IVertex)graphNode).InArcs;
      if (inArcs.Count > 0) {
        if (inArcs.Count == 1) {
          genealogyGraphChart.SelectedGraphNode = (IGenealogyGraphNode<ISymbolicExpressionTree>)inArcs[0].Source;
          return;
        }
        IGenealogyGraphNode<ISymbolicExpressionTree> source = null;
        var fragment = ((IArc<IDeepCloneable>)inArcs.Last()).Data as IFragment<ISymbolicExpressionTreeNode>;
        var traceData = ((IArc<IDeepCloneable>)inArcs[0]).Data as TraceData;
        if (fragment != null) {
          source = (IGenealogyGraphNode<ISymbolicExpressionTree>)inArcs.Last().Source;
        }
        if (traceData != null) {
          var f = graphNode.Data.NodeAt(traceData.LastFragmentIndex);
          var selected = inArcs.Skip(1).First(x => {
            var td = (TraceData)((IArc<IDeepCloneable>)x).Data;
            var s = (IGenealogyGraphNode<ISymbolicExpressionTree>)x.Source;
            return f.Difference(s.Data.NodeAt(td.SubtreeIndex)) == null;
          });
          source = (IGenealogyGraphNode<ISymbolicExpressionTree>)selected.Source;
        }
        if (source != null)
          genealogyGraphChart.SelectedGraphNode = source;
      }
    }
    #endregion
  }

  // bogus class needed in order to be able to "design" the view
  public class SymbolicDataAnalysisGenealogyGraphViewDesignable : GenealogyGraphView<ISymbolicExpressionTree> { }

  #region helper methods
  internal static class Util {
    internal static ISymbolicExpressionTreeNode NodeAt(this ISymbolicExpressionTree tree, int position) {
      return NodeAt(tree.Root, position);
    }
    internal static ISymbolicExpressionTreeNode NodeAt(this ISymbolicExpressionTreeNode root, int position) {
      return root.IterateNodesPrefix().ElementAt(position);
    }
  }
  #endregion
}
